@@ -15,10 +15,24 @@
 #include <fcntl.h>
 #include <errno.h>
 
+#ifdef LINC_SSL_SUPPORT
+#include <openssl/ssl.h>
+#endif
+
+#include "linc-debug.h"
 #include "linc-private.h"
 #include <linc/linc-config.h>
 #include <linc/linc-connection.h>
 
+struct _LINCConnectionPrivate {
+#ifdef LINC_SSL_SUPPORT
+	SSL        *ssl;
+#endif
+	GIOChannel *gioc;
+	LincWatch  *tag;
+	int         fd;
+};
+
 static GObjectClass *parent_class = NULL;
 
 enum {
@@ -27,24 +41,44 @@ enum {
 };
 static guint linc_connection_signals [LAST_SIGNAL];
 
+
+static gboolean linc_connection_connected (GIOChannel  *gioc,
+					   GIOCondition condition,
+					   gpointer     data);
+
 static void
-linc_source_remove (LINCConnection *cnx, gboolean unref)
+linc_source_remove (LINCConnection *cnx)
 {
-	if (cnx->tag) {
-		LincWatch *thewatch = cnx->tag;
-		cnx->tag = NULL;
+	if (cnx->priv->tag) {
+		LincWatch *thewatch = cnx->priv->tag;
+		cnx->priv->tag = NULL;
 		linc_io_remove_watch (thewatch);
-		if (unref) 
-			g_object_unref (G_OBJECT (cnx));
+		d_printf ("Removed watch on %d\n", cnx->priv->fd);
 	}
 }
 
+static void
+linc_source_add (LINCConnection *cnx,
+		 GIOCondition    condition)
+{
+	g_assert (cnx->priv->tag == NULL);
+
+	cnx->priv->tag = linc_io_add_watch (
+		cnx->priv->gioc, condition,
+		linc_connection_connected, cnx);
+
+	d_printf ("Added watch on %d (0x%x)\n",
+		 cnx->priv->fd, condition);
+}
+
 static void
 linc_close_fd (LINCConnection *cnx)
 {
-	if (cnx->fd >= 0)
-		close (cnx->fd);
-	cnx->fd = -1;
+	if (cnx->priv->fd >= 0) {
+		d_printf ("Close %d\n", cnx->priv->fd);
+		close (cnx->priv->fd);
+	}
+	cnx->priv->fd = -1;
 }
 
 static void
@@ -52,26 +86,43 @@ linc_connection_dispose (GObject *obj)
 {
 	LINCConnection *cnx = (LINCConnection *)obj;
 
-	linc_source_remove (cnx, FALSE);
+	linc_source_remove (cnx);
 
-	g_free (cnx->remote_host_info);
-	cnx->remote_host_info = NULL;
+	if (cnx->priv->gioc) {
+		g_io_channel_unref (cnx->priv->gioc);
+		cnx->priv->gioc = NULL;
+	}
 
-	g_free (cnx->remote_serv_info);
-	cnx->remote_serv_info = NULL;
+	parent_class->dispose (obj);
+}
 
-	if (cnx->gioc)
-		g_io_channel_unref (cnx->gioc);
-	cnx->gioc = NULL;
+static void
+linc_connection_finalize (GObject *obj)
+{
+	LINCConnection *cnx = (LINCConnection *)obj;
 
 	linc_close_fd (cnx);
 
-	if (parent_class->dispose)
-		parent_class->dispose (obj);
+	g_free (cnx->remote_host_info);
+	g_free (cnx->remote_serv_info);
+
+	g_free (cnx->priv);
+
+	parent_class->finalize (obj);
 }
 
-#define ERR_CONDS (G_IO_ERR|G_IO_HUP|G_IO_NVAL)
-#define IN_CONDS  (G_IO_PRI|G_IO_IN)
+
+/*
+ * FIXME: a horribly inefficient way to simply stop
+ * listening for G_IO_OUT - which we use initialy to
+ * be notified of connection.
+ */
+static void
+linc_source_unwatch_out (LINCConnection *cnx)
+{
+	linc_source_remove (cnx);
+	linc_source_add (cnx, LINC_ERR_CONDS | LINC_IN_CONDS);
+}
 
 static gboolean
 linc_connection_connected (GIOChannel  *gioc,
@@ -82,45 +133,46 @@ linc_connection_connected (GIOChannel  *gioc,
 	LINCConnectionClass *klass;
 	int rv, n;
 	socklen_t n_size = sizeof(n);
-	gboolean disconnected = FALSE;
 
 	g_object_ref (G_OBJECT (cnx));
 
 	klass = LINC_CONNECTION_CLASS (G_OBJECT_GET_CLASS (data));
 
 	if (cnx->status == LINC_CONNECTED &&
-	    (condition & IN_CONDS) &&
-	    klass->handle_input)
+	    (condition & LINC_IN_CONDS) &&
+	    klass->handle_input) {
 
+		d_printf ("Handle input on fd %d\n", cnx->priv->fd);
 		klass->handle_input (cnx);
 
-	else if (condition & (ERR_CONDS | G_IO_OUT)) {
+	} else if (condition & (LINC_ERR_CONDS | G_IO_OUT)) {
 
 		switch (cnx->status) {
 		case LINC_CONNECTING:
 			n = 0;
-			rv = getsockopt (cnx->fd, SOL_SOCKET, SO_ERROR, &n, &n_size);
+			rv = getsockopt (cnx->priv->fd, SOL_SOCKET, SO_ERROR, &n, &n_size);
 			if (!rv && !n && condition == G_IO_OUT) {
+				d_printf ("State changed to connected on %d\n", cnx->priv->fd);
+				linc_source_unwatch_out (cnx);
 				linc_connection_state_changed (cnx, LINC_CONNECTED);
-				g_assert (cnx->tag == NULL);
-				cnx->tag = linc_io_add_watch (
-					cnx->gioc, ERR_CONDS,
-					linc_connection_connected, cnx);
-
+				
 			} else {
+				d_printf ("Error connecting %d %d %d on fd %d\n",
+					   rv, n, errno, cnx->priv->fd);
 				linc_connection_state_changed (cnx, LINC_DISCONNECTED);
-				disconnected = TRUE;
 			}
 			break;
 		case LINC_CONNECTED: {
-			linc_connection_state_changed(cnx, LINC_DISCONNECTED);
-			disconnected = TRUE;
+			d_printf ("Disconnect %d\n", cnx->priv->fd);
+			linc_connection_state_changed (cnx, LINC_DISCONNECTED);
 			break;
 		}
 		default:
 			break;
 		}
-	}
+	} else
+		g_warning ("Dropping state on fd %d on the floor (%d)",
+			   cnx->priv->fd, condition);
 
 	g_object_unref (G_OBJECT (cnx));
 
@@ -147,33 +199,33 @@ static void
 linc_connection_class_state_changed (LINCConnection      *cnx,
 				     LINCConnectionStatus status)
 {
+	d_printf ("State changing from '%s' to '%s' on fd %d\n",
+		 STATE_NAME (cnx->status), STATE_NAME (status),
+		 cnx->priv->fd);
+
 	switch (status) {
 	case LINC_CONNECTED:
 #ifdef LINC_SSL_SUPPORT
 		if (cnx->options & LINC_CONNECTION_SSL) {
 			if (cnx->was_initiated)
-				SSL_connect (cnx->ssl);
+				SSL_connect (cnx->priv->ssl);
 			else
-				SSL_accept (cnx->ssl);
+				SSL_accept (cnx->priv->ssl);
 		}
 #endif
-		if (!cnx->tag)
-			cnx->tag = linc_io_add_watch (
-				cnx->gioc, ERR_CONDS | IN_CONDS,
-				linc_connection_connected, cnx);
+		if (!cnx->priv->tag)
+			linc_source_add (cnx, LINC_ERR_CONDS | LINC_IN_CONDS);
 		break;
 
 	case LINC_CONNECTING:
-		linc_source_remove (cnx, FALSE);
-		/* FIXME: this looks bogus (?) */
-		g_assert (cnx->tag == NULL);
-		cnx->tag = linc_io_add_watch (
-			cnx->gioc, G_IO_OUT|ERR_CONDS,
-			linc_connection_connected, cnx);
+		/* FIXME: We might be re-connecting, and need to watch
+		 * G_IO_OUT - again this could be more efficient */
+		linc_source_remove (cnx);
+		linc_source_add (cnx, G_IO_OUT | LINC_ERR_CONDS);
 		break;
 
 	case LINC_DISCONNECTED:
-		linc_source_remove (cnx, TRUE);
+		linc_source_remove (cnx);
 		linc_close_fd (cnx);
 
 		g_signal_emit (G_OBJECT (cnx),
@@ -189,8 +241,8 @@ linc_connection_class_state_changed (LINCConnection      *cnx,
  * @cnx: a #LINCConnection.
  * @fd: a connected/connecting file descriptor.
  * @proto: a #LINCProtocolInfo.
- * @host: protocol dependant host information.
- * @service: protocol dependant service information(e.g. port number).
+ * @remote_host_info: protocol dependant host information; gallocation swallowed
+ * @remote_serv_info: protocol dependant service information(e.g. port number). gallocation swallowed
  * @was_initiated: #TRUE if the connection was initiated by us.
  * @status: a #LINCConnectionStatus value.
  * @options: combination of #LINCConnectionOptions.
@@ -204,27 +256,34 @@ gboolean
 linc_connection_from_fd (LINCConnection         *cnx,
 			 int                     fd,
 			 const LINCProtocolInfo *proto,
-			 char                   *host,
-			 char                   *service,
+			 gchar                  *remote_host_info,
+			 gchar                  *remote_serv_info,
 			 gboolean                was_initiated,
 			 LINCConnectionStatus    status,
 			 LINCConnectionOptions   options)
 {
-	cnx->fd               = fd;
-	cnx->gioc             = g_io_channel_unix_new (fd);
-	cnx->was_initiated    = was_initiated;
-	cnx->is_auth          = (proto->flags & LINC_PROTOCOL_SECURE) ? TRUE : FALSE;
-	cnx->remote_host_info = host;
-	cnx->remote_serv_info = service;
-	cnx->options          = options;
+	cnx->was_initiated = was_initiated;
+	cnx->is_auth       = (proto->flags & LINC_PROTOCOL_SECURE);
+	cnx->proto         = proto;
+	cnx->options       = options;
+	cnx->priv->fd      = fd;
+	cnx->priv->gioc    = g_io_channel_unix_new (fd);
+
+	cnx->remote_host_info = remote_host_info;
+	cnx->remote_serv_info = remote_serv_info;
+
+	d_printf ("Cnx from fd (%d) '%s', '%s', '%s'\n",
+		 fd, proto->name, 
+		 remote_host_info ? remote_host_info : "<Null>",
+		 remote_serv_info ? remote_serv_info : "<Null>");
 
 	if (proto->setup)
 		proto->setup (fd, options);
 
 #ifdef LINC_SSL_SUPPORT
 	if (options & LINC_CONNECTION_SSL) {
-		cnx->ssl = SSL_new (linc_ssl_ctx);
-		SSL_set_fd (cnx->ssl, fd);
+		cnx->priv->ssl = SSL_new (linc_ssl_ctx);
+		SSL_set_fd (cnx->priv->ssl, fd);
 	}
 #endif
 
@@ -268,8 +327,9 @@ linc_connection_initiate (LINCConnection        *cnx,
 		return FALSE;
 
 
-	saddr = linc_protocol_get_sockaddr (proto, host, 
-					    service, &saddr_len);
+	saddr = linc_protocol_get_sockaddr (
+		proto, host, service, &saddr_len);
+
 	if (!saddr)
 		return FALSE;
 
@@ -290,16 +350,20 @@ linc_connection_initiate (LINCConnection        *cnx,
 	if (rv && errno != EINPROGRESS)
 		goto out;
 
+	d_printf ("initiate 'connect' on new fd %d [ %d; %d ]\n",
+		 fd, rv, errno);
+
 	retval = linc_connection_from_fd (
-				cnx, fd, proto, 
-				g_strdup (host),
-				g_strdup (service),
-				TRUE, rv ? LINC_CONNECTING : LINC_CONNECTED,
-				options);
+		cnx, fd, proto, 
+		g_strdup (host), g_strdup (service),
+		TRUE, rv ? LINC_CONNECTING : LINC_CONNECTED,
+		options);
 
  out:
-	if (!cnx && fd >= 0)
+	if (!cnx && fd >= 0) {
+		d_printf ("initiation failed\n");
 		close (fd);
+	}
 
 	g_free (saddr);
 
@@ -322,7 +386,7 @@ linc_connection_state_changed (LINCConnection      *cnx,
 	klass = (LINCConnectionClass *)G_OBJECT_GET_CLASS (cnx);
 
 	if (klass->state_changed)
-		klass->state_changed(cnx, status);
+		klass->state_changed (cnx, status);
 }
 
 /**
@@ -332,11 +396,7 @@ linc_connection_state_changed (LINCConnection      *cnx,
  * @len: the length of the array in bytes to read ingo
  * @block_for_full_read: whether to block for a full read
  * 
- * FIXME: it allows re-enterancy via linc_connection_iterate
- *        in certain cases.
- * FIXME: on this basis, the connection can die underneath
- *        our feet eg. between the main_iteration and the
- *        g_return_if_fail.
+ * Warning, block_for_full_read is of limited usefullness.
  *
  * Return value: number of bytes written on success; negative on error.
  **/
@@ -348,14 +408,11 @@ linc_connection_read (LINCConnection *cnx,
 {
 	int bytes_read = 0;
 
+	d_printf ("Read up to %d bytes from fd %d\n", len, cnx->priv->fd);
+
 	if (!len)
 		return 0;
 
-	if (cnx->options & LINC_CONNECTION_NONBLOCKING) {
-		while (cnx->status == LINC_CONNECTING)
-			linc_main_iteration (TRUE);
-	}
-
 	if (cnx->status != LINC_CONNECTED)
 		return -1;
 
@@ -364,20 +421,22 @@ linc_connection_read (LINCConnection *cnx,
 
 #ifdef LINC_SSL_SUPPORT
 		if (cnx->options & LINC_CONNECTION_SSL)
-			n = SSL_read (cnx->ssl, buf, len);
+			n = SSL_read (cnx->priv->ssl, buf, len);
 		else
 #endif
-			n = read (cnx->fd, buf, len);
-		
+			n = read (cnx->priv->fd, buf, len);
+
 		if (n < 0) {
 #ifdef LINC_SSL_SUPPORT
 			if (cnx->options & LINC_CONNECTION_SSL) {
 				gulong rv;
-				rv = SSL_get_error (cnx->ssl, n);
-				if ((rv == SSL_ERROR_WANT_READ
-				     || rv == SSL_ERROR_WANT_WRITE)
-				    && (cnx->options & LINC_CONNECTION_NONBLOCKING))
-					linc_main_iteration (FALSE);
+
+				rv = SSL_get_error (cnx->priv->ssl, n);
+
+				if ((rv == SSL_ERROR_WANT_READ ||
+				     rv == SSL_ERROR_WANT_WRITE) &&
+				    (cnx->options & LINC_CONNECTION_NONBLOCKING))
+					return bytes_read;
 				else
 					return -1;
 			} else
@@ -386,23 +445,30 @@ linc_connection_read (LINCConnection *cnx,
 				if (errno == EINTR)
 					continue;
 
-				if (errno == EAGAIN
-				   && (cnx->options & LINC_CONNECTION_NONBLOCKING))
-					linc_main_iteration (FALSE);
-				else
+				else if (errno == EAGAIN &&
+					 (cnx->options & LINC_CONNECTION_NONBLOCKING))
+					return bytes_read;
+
+				else if (errno == EBADF) {
+					g_warning ("Serious fd usage error %d", cnx->priv->fd);
+					return -1;
+
+				} else
 					return -1;
 			}
 
-		} else if (n == 0)
-			return 0;
-
-		else {
+		} else if (n == 0) {
+			if (cnx->options & LINC_CONNECTION_NONBLOCKING)
+				return -1;
+		} else {
 			buf += n;
 			len -= n;
 			bytes_read += n;
 		}
 	} while (len > 0 && block_for_full_read);
 
+	d_printf ("we read %d bytes\n", bytes_read);
+
 	return bytes_read;
 }
 
@@ -436,20 +502,22 @@ linc_connection_write (LINCConnection *cnx,
 
 	g_return_val_if_fail (cnx->status == LINC_CONNECTED, -1);
 
+	d_printf ("Write %ld bytes to fd %d\n", len, cnx->priv->fd);
+
 	while (len > 0) {
 #ifdef LINC_SSL_SUPPORT
 		if (cnx->options & LINC_CONNECTION_SSL)
-			n = SSL_write (cnx->ssl, buf, len);
+			n = SSL_write (cnx->priv->ssl, buf, len);
 		else
 #endif
-			n = write (cnx->fd, buf, len);
+			n = write (cnx->priv->fd, buf, len);
 
 		if (n < 0) {
 #ifdef LINC_SSL_SUPPORT
 			if (cnx->options & LINC_CONNECTION_SSL) {
 				gulong rv;
 
-				rv = SSL_get_error (cnx->ssl, n);
+				rv = SSL_get_error (cnx->priv->ssl, n);
 
 				if ((rv == SSL_ERROR_WANT_READ || 
 				     rv == SSL_ERROR_WANT_WRITE) &&
@@ -467,7 +535,11 @@ linc_connection_write (LINCConnection *cnx,
 					 (cnx->options & LINC_CONNECTION_NONBLOCKING))
 					linc_main_iteration (FALSE);
 
-				else
+				else if (errno == EBADF) {
+					g_warning ("Serious fd usage error %d", cnx->priv->fd);
+					return -1;
+
+				} else
 					return -1;
 			}
 		} else if (n == 0)
@@ -508,6 +580,8 @@ linc_connection_writev (LINCConnection *cnx,
 			linc_main_iteration (TRUE);
 	}
 
+	d_printf ("Writev %ld bytes to fd %d\n", total_size, cnx->priv->fd);
+
 	g_return_val_if_fail (cnx->status == LINC_CONNECTED, -1);
 
 #ifdef LINC_SSL_SUPPORT
@@ -531,7 +605,7 @@ linc_connection_writev (LINCConnection *cnx,
 		vecs_left = nvecs;
 		size_left = total_size;
 
-		fd = cnx->fd;
+		fd = cnx->priv->fd;
 
 		while (size_left > 0) {
 			int n;
@@ -545,7 +619,11 @@ linc_connection_writev (LINCConnection *cnx,
 				else if (errno == EAGAIN &&
 					 (cnx->options & LINC_CONNECTION_NONBLOCKING))
 					linc_main_iteration (FALSE);
-				else
+
+				else if (errno == EBADF) {
+					g_warning ("Serious fd usage error %d", cnx->priv->fd);
+					return -1;
+				} else
 					return -1; /* Unhandlable error */
 
 			} else if (n == 0)
@@ -571,7 +649,8 @@ linc_connection_writev (LINCConnection *cnx,
 static void
 linc_connection_init (LINCConnection *cnx)
 {
-	cnx->fd = -1;
+	cnx->priv = g_new0 (LINCConnectionPrivate, 1);
+	cnx->priv->fd = -1;
 }
 
 static void
@@ -579,7 +658,8 @@ linc_connection_class_init (LINCConnectionClass *klass)
 {
 	GObjectClass *object_class = (GObjectClass *) klass;
 
-	object_class->dispose = linc_connection_dispose;
+	object_class->dispose  = linc_connection_dispose;
+	object_class->finalize = linc_connection_finalize;
 
 	klass->state_changed  = linc_connection_class_state_changed;
 	klass->broken         = NULL;
