@@ -14,15 +14,22 @@
 #include <netdb.h>
 #include <unistd.h>
 #include <string.h>
+#include <fcntl.h>
 #include <errno.h>
 
 #include <linc/linc.h>
 #include <linc/linc-server.h>
 #include <linc/linc-connection.h>
 
+#include "linc-debug.h"
 #include "linc-private.h"
 
-#undef DEBUG
+struct _LINCServerPrivate {
+	int        fd;
+	GMutex    *mutex;
+	LincWatch *tag;
+	GSList    *connections;
+};
 
 enum {
 	NEW_CONNECTION,
@@ -67,41 +74,61 @@ my_cclosure_marshal_VOID__OBJECT (GClosure     *closure,
 static void
 linc_server_init (LINCServer *cnx)
 {
-	cnx->mutex = linc_mutex_new ();
-	cnx->fd = -1;
+	cnx->priv = g_new0 (LINCServerPrivate, 1);
+
+	cnx->priv->mutex = linc_mutex_new ();
+	cnx->priv->fd = -1;
 }
 
 static void
 linc_server_dispose (GObject *obj)
 {
+	GSList     *l;
 	LINCServer *cnx = (LINCServer *)obj;
 
+	d_printf ("Dispose / close server fd %d", cnx->priv->fd);
+
 #ifdef G_THREADS_ENABLED
-	if (cnx->mutex)
-		g_mutex_free (cnx->mutex);
+	if (cnx->priv->mutex) {
+		g_mutex_free (cnx->priv->mutex);
+		cnx->priv->mutex = NULL;
+	}
 #endif
-	if (cnx->tag)
-		linc_io_remove_watch (cnx->tag);
-	cnx->tag = NULL;
+	if (cnx->priv->tag) {
+		linc_io_remove_watch (cnx->priv->tag);
+		cnx->priv->tag = NULL;
+	}
 
-	if (cnx->proto && cnx->proto->destroy)
-		cnx->proto->destroy (
-			cnx->fd, cnx->local_host_info,
-			cnx->local_serv_info);
-	cnx->proto = NULL;
+	if (cnx->priv->fd >= 0) {
+		if (cnx->proto->destroy)
+			cnx->proto->destroy (
+				cnx->priv->fd, cnx->local_host_info,
+				cnx->local_serv_info);
+		close (cnx->priv->fd);
+		cnx->priv->fd = -1;
+	}
 
-	if (cnx->fd >= 0)
-		close (cnx->fd);
-	cnx->fd = -1;
+	for (; (l = cnx->priv->connections); ) {
+		GObject *o = l->data;
+		cnx->priv->connections = l->next;
+		g_slist_free_1 (l);
+		g_object_unref (o);
+	}
+
+	parent_class->dispose (obj);
+}
+
+static void
+linc_server_finalize (GObject *obj)
+{
+	LINCServer *cnx = (LINCServer *)obj;
 
 	g_free (cnx->local_host_info);
-	cnx->local_host_info = NULL;
-
 	g_free (cnx->local_serv_info);
-	cnx->local_serv_info = NULL;
 
-	if (parent_class->dispose)
-		parent_class->dispose (obj);
+	g_free (cnx->priv);
+
+	parent_class->finalize (obj);
 }
 
 static LINCConnection *
@@ -111,24 +138,33 @@ linc_server_create_connection (LINCServer *cnx)
 }
 
 static gboolean
-linc_server_accept_connection (LINCServer *server, LINCConnection **connection)
+linc_server_accept_connection (LINCServer      *server,
+			       LINCConnection **connection)
 {
 	LINCServerClass *klass;
 	struct sockaddr *saddr;
 	int              addrlen, fd;
-	char            *hostname;
-	char            *service;
 	
 	g_return_val_if_fail (connection != NULL, FALSE);
 
 	addrlen = server->proto->addr_len;
 	saddr = g_alloca (addrlen);
 
-	fd = accept (server->fd, saddr, &addrlen);
-	if (fd < 0)
+	fd = accept (server->priv->fd, saddr, &addrlen);
+	if (fd < 0) {
+		d_printf ("accept on %d failed %d", server->priv->fd, errno);
 		return FALSE; /* error */
+	}
 
-	if (!linc_protocol_get_sockinfo (server->proto, saddr, &hostname, &service)) {
+	if (server->create_options & LINC_CONNECTION_NONBLOCKING)
+		if (fcntl (fd, F_SETFL, O_NONBLOCK) < 0) {
+			d_printf ("failed to set O_NONBLOCK on %d", fd);
+			close (fd);
+			return FALSE;
+		}
+
+	if (fcntl (fd, F_SETFD, FD_CLOEXEC) < 0) {
+		d_printf ("failed to set cloexec on %d", fd);
 		close (fd);
 		return FALSE;
 	}
@@ -140,19 +176,22 @@ linc_server_accept_connection (LINCServer *server, LINCConnection **connection)
 
 	g_return_val_if_fail (*connection != NULL, FALSE);
 
-	if (!linc_connection_from_fd (
-		*connection, fd, server->proto, hostname, service,
-		FALSE, LINC_CONNECTED, server->create_options)) {
+	d_printf ("accepted a new connection (%d) on server %d\n",
+		 fd, server->priv->fd);
 
+	if (!linc_connection_from_fd (
+		*connection, fd, server->proto, NULL, NULL,
+		FALSE, LINC_CONNECTED, server->create_options)) {
+		
 		g_object_unref (G_OBJECT (*connection));
 		*connection = NULL;
 
-		g_free (hostname);
-		g_free (service);
-
 		return FALSE;
 	}
 
+	server->priv->connections = g_slist_prepend (
+		server->priv->connections, *connection);
+
 	return TRUE;
 }
 
@@ -165,14 +204,14 @@ linc_server_handle_io (GIOChannel  *gioc,
 	LINCServer     *server = data;
 	LINCConnection *connection;
 
-	if (condition != G_IO_IN)
-		g_error ("condition on server fd is %#x", condition);
+	if (!(condition & LINC_IN_CONDS))
+		g_error ("error condition on server fd is %#x", condition);
 
-	LINC_MUTEX_LOCK (server->mutex);
+	LINC_MUTEX_LOCK (server->priv->mutex);
 
 	accepted = linc_server_accept_connection (server, &connection);
 
-	LINC_MUTEX_UNLOCK (server->mutex);
+	LINC_MUTEX_UNLOCK (server->priv->mutex);
 
 	if (!accepted) {
 		GValue parms[2];
@@ -192,12 +231,6 @@ linc_server_handle_io (GIOChannel  *gioc,
 	return TRUE;
 }
 
-void
-linc_server_handle (LINCServer *cnx)
-{
-	linc_server_handle_io (NULL, G_IO_IN, cnx);
-}
-
 gboolean
 linc_server_setup (LINCServer            *cnx,
 		   const char            *proto_name,
@@ -220,9 +253,7 @@ linc_server_setup (LINCServer            *cnx,
 
 	proto = linc_protocol_find (proto_name);
 	if (!proto) {
-#ifdef DEBUG
-		fprintf (stderr, "Can't find proto '%s'\n", proto_name);
-#endif
+		d_printf ("Can't find proto '%s'\n", proto_name);
 		return FALSE;
 	}
 
@@ -238,11 +269,9 @@ linc_server_setup (LINCServer            *cnx,
 					    local_serv_info, &saddr_len);
 
 	if (!saddr) {
-#ifdef DEBUG
-		fprintf (stderr, "Can't get_sockaddr proto '%s' '%s'\n",
+		d_printf ("Can't get_sockaddr proto '%s' '%s'\n",
 			 local_host_info,
 			 local_serv_info ? local_serv_info : "(null)");
-#endif
 		return FALSE;
 	}
 
@@ -250,11 +279,9 @@ linc_server_setup (LINCServer            *cnx,
 		     proto->stream_proto_num);
 	if (fd < 0) {
 		g_free (saddr);
-#ifdef DEBUG
-		fprintf (stderr, "socket (%d, %d, %d) failed\n",
+		d_printf ("socket (%d, %d, %d) failed\n",
 			 proto->family, SOCK_STREAM, 
 			 proto->stream_proto_num);
-#endif
 		return FALSE;
 	}
 
@@ -271,71 +298,59 @@ linc_server_setup (LINCServer            *cnx,
 		n = bind (fd, saddr, saddr_len);
 
 	if (n && errno == EADDRINUSE) {
-#ifdef DEBUG
-		perror ("bind failed, retrying.");
-#endif
+		d_printf ("bind failed; retrying");
 		goto address_in_use;
 	}
 
 	if (!n)
 		n = listen (fd, 10);
-#ifdef DEBUG
-	else {
-		fprintf (stderr, "Errno: %d (0x%x)\n", errno, errno);
-		perror ("bind really failed");
-	}
-#endif
-
-	/*
-	 * FIXME: If we get EINUSE we should loop to getaddrinfo ?
-	 */
+	else
+		d_printf ("bind really failed errno: %d\n", errno);
 
 	if (!n)
 		n = getsockname (fd, saddr, &saddr_len);
-#ifdef DEBUG
 	else
-		perror ("listen failed");
-#endif
+		d_printf ("listen failed errno: %d\n", errno);
 
 	if (n) {
-		close(fd);
-#ifdef DEBUG
-		perror ("get_sockname failed");
-#endif
+		close (fd);
+		d_printf ("get_sockname failed errno: %d\n", errno);
 		return FALSE;
 	}
 
 	if (!linc_protocol_get_sockinfo (proto, saddr, &hostname, &service)) {
 
 		g_free (saddr);
-#ifdef DEBUG
-		fprintf (stderr, "linc_getsockinfo failed.\n");
-#endif
+		d_printf ("linc_getsockinfo failed.\n");
 		return FALSE;
 	}
 
 	g_free (saddr);
 
 	cnx->proto = proto;
-	cnx->fd = fd;
+	cnx->priv->fd = fd;
 
 	if ((create_options & LINC_CONNECTION_NONBLOCKING)) {
 		gioc = g_io_channel_unix_new (fd);
 
-		g_assert (cnx->tag == NULL);
+		g_assert (cnx->priv->tag == NULL);
 
-		cnx->tag = linc_io_add_watch (
-			gioc, 
-			G_IO_IN|G_IO_HUP|G_IO_ERR|G_IO_NVAL,
+		cnx->priv->tag = linc_io_add_watch (
+			gioc, LINC_IN_CONDS | LINC_ERR_CONDS,
 			linc_server_handle_io, cnx);
 
-		g_io_channel_unref(gioc);
+		g_io_channel_unref (gioc);
 	}
 
-	cnx->create_options = create_options;
+	cnx->create_options  = create_options;
 	cnx->local_host_info = hostname;
 	cnx->local_serv_info = service;
 
+	d_printf ("Created a new server fd (%d) '%s', '%s', '%s'\n",
+		 fd, proto->name, 
+		 hostname ? hostname : "<Null>",
+		 service ? service : "<Null>");
+
 	return TRUE;
 }
 
@@ -347,6 +362,7 @@ linc_server_class_init (LINCServerClass *klass)
 	GObjectClass *object_class = (GObjectClass *) klass;
 
 	object_class->dispose    = linc_server_dispose;
+	object_class->finalize   = linc_server_finalize;
 	klass->create_connection = linc_server_create_connection;
 
 	parent_class = g_type_class_peek_parent (klass);
