@@ -10,14 +10,14 @@
  *                 Sun Microsystems, Inc.
  */
 #include <signal.h>
+#include "linc-debug.h"
 #include "linc-private.h"
 
-#undef WATCH_DEBUG
-
 static gboolean linc_threaded = FALSE;
 static gboolean linc_mutex_new_called = FALSE;
 GMainLoop      *linc_loop = NULL;
 GMainContext   *linc_context = NULL;
+static GMutex  *linc_lifecycle_mutex = NULL;
 
 #ifdef LINC_SSL_SUPPORT
 SSL_METHOD *linc_ssl_method;
@@ -108,6 +108,8 @@ linc_init (gboolean init_threads)
 	linc_ssl_method = SSLv23_method ();
 	linc_ssl_ctx = SSL_CTX_new (linc_ssl_method);
 #endif
+
+	linc_lifecycle_mutex = linc_mutex_new ();
 }
 
 struct _LincWatch {
@@ -119,6 +121,19 @@ struct _LincWatch {
 static GSList *watches = NULL;
 #endif
 
+/**
+ * linc_io_add_watch:
+ * @channel: the GIOChannel to watch
+ * @condition: the condition mask to watch for
+ * @func: the function to invoke when a condition is met
+ * @user_data: a user data closure
+ * 
+ * This routine creates a watch on an IO channel that operates both in
+ * the standard glib mainloop, but also in the 'linc' mainloop so we
+ * can iterate that without causing re-enterancy.
+ * 
+ * Return value: a pointer identifying the watch.
+ **/
 LincWatch *
 linc_io_add_watch (GIOChannel    *channel,
 		   GIOCondition   condition,
@@ -147,12 +162,28 @@ linc_io_add_watch (GIOChannel    *channel,
 	g_source_unref (source);
 
 #ifdef WATCH_DEBUG
+	source = g_main_context_find_source_by_id (
+		NULL, w->main_id);
+	g_assert (source != NULL);
+	g_assert (source->poll_fds != NULL);
+
+	source = g_main_context_find_source_by_id (
+		linc_context, w->linc_id);
+	g_assert (source != NULL);
+	g_assert (source->poll_fds != NULL);
+
 	watches = g_slist_prepend (watches, w);
 #endif
 	
 	return w;
 }
 
+/**
+ * linc_io_remove_watch:
+ * @watch: the handle of a watch on a GIOChannel
+ * 
+ * This removes a watch by it's handle in @w
+ **/
 void
 linc_io_remove_watch (LincWatch *w)
 {
@@ -184,6 +215,13 @@ linc_io_remove_watch (LincWatch *w)
 	}
 }
 
+/**
+ * linc_main_iteration:
+ * @block_for_reply: whether we should wait for a reply
+ * 
+ * This routine iterates the linc mainloop, which has
+ * only the linc sources registered against it.
+ **/
 void
 linc_main_iteration (gboolean block_for_reply)
 {
@@ -210,18 +248,38 @@ linc_main_iteration (gboolean block_for_reply)
 		linc_context, block_for_reply);
 }
 
+/**
+ * linc_main_pending:
+ * 
+ * determines if the linc mainloop has any pending work to process.
+ * 
+ * Return value: TRUE if the linc mainloop has any pending work to process.
+ **/
 gboolean
 linc_main_pending (void)
 {
 	return g_main_context_pending (linc_context);
 }
 
+/**
+ * linc_main_loop_run:
+ * 
+ * Runs the linc mainloop; blocking until the loop is exited.
+ **/
 void
 linc_main_loop_run (void)
 {
 	g_main_loop_run (linc_loop);
 }
 
+/**
+ * linc_mutex_new:
+ * 
+ * Creates a mutes, iff threads are supported, initialized and
+ * linc_set_threaded has been called.
+ * 
+ * Return value: a new GMutex, or NULL if one is not required.
+ **/
 GMutex *
 linc_mutex_new (void)
 {
@@ -235,3 +293,42 @@ linc_mutex_new (void)
 	return NULL;
 }
 
+GMutex *
+linc_object_get_mutex (void)
+{
+	return linc_lifecycle_mutex;
+}
+
+gpointer
+linc_object_ref (GObject *object)
+{
+	gpointer ret;
+
+	LINC_MUTEX_LOCK   (linc_lifecycle_mutex);
+
+	ret = g_object_ref (object);
+
+	LINC_MUTEX_UNLOCK (linc_lifecycle_mutex);
+
+	return ret;
+}
+
+void
+linc_object_unref (GObject *object)
+{
+	LINC_MUTEX_LOCK   (linc_lifecycle_mutex);
+
+	g_object_unref (object);
+
+	/* FIXME: we need to do our own ref counting here
+	 * so finalization doesn't happen whilst holding
+	 * this lock ! can't use these methods until then */
+
+	LINC_MUTEX_UNLOCK (linc_lifecycle_mutex);
+}
+
+GMainLoop *
+linc_main_get_loop (void)
+{
+	return linc_loop;
+}
